// OpenGL ES 3.1 Raytracer fragment shader.
// Written by Nils Liaaen Corneliusen.
// Please refer to the article at http://www.ignorantus.com/pages/gpu_hacks/ for more information.
// Based on code by Shodruky Rhyammer: https://github.com/shodruky-rhyammer/blobubska
#version 310 es
precision highp float;
layout(location = 0) in vec2 v_texCoord;
layout(location = 0) out vec4 outColor;

#define OBJNUM (5*16)
#define REFLECTIONS 3

#define EPSILON 0.0001f
#define FARDIST 256.0f

//#pragma optionNV(unroll all)

layout(binding = 0) uniform stuff
{
    vec4 obj_pos[OBJNUM];
    vec4 obj_col[OBJNUM];
    vec4 plane_norm;
    vec4 plane_col;
    vec4 eye;
    vec4 light_pos;
    vec4 light_col;
    vec4 startpos;
    float ax;
    float ay;
    float frame;
    float fx_offset;
};


void main()
{
    vec3 ray_dir = normalize( vec3( startpos.x + ax * v_texCoord.x,
                                    startpos.y + ay * v_texCoord.y,
                                    startpos.z ) );

    vec3 col = vec3( 0.0f );

    vec3 ray_pos = eye.xyz;

    // wobble
    if( fx_offset != 0.0f ) {
        ray_pos.x += sin( v_texCoord.x*4.0f + obj_pos[1].x ) * fx_offset;
        ray_pos.y += sin( v_texCoord.y*2.0f                ) * fx_offset;
        ray_pos.z += sin( v_texCoord.x*4.0f + obj_pos[1].z ) * fx_offset;
    }

    for( int refl = 0; refl < REFLECTIONS; refl++ ) {

        float rt_dist = FARDIST;
        int rt_hit = -1;

        // trace_ray()
        for( int i = 0; i < OBJNUM; i += 16 ) {
            vec3 v0, v1, v2, v3;
            vec4 b03, d03;
            bvec4 dv;

            // 0-3
            v0 = ray_pos - obj_pos[i+0].xyz;
            v1 = ray_pos - obj_pos[i+1].xyz;
            v2 = ray_pos - obj_pos[i+2].xyz;
            v3 = ray_pos - obj_pos[i+3].xyz;

            b03 = vec4( dot( v0, ray_dir ),
                        dot( v1, ray_dir ),
                        dot( v2, ray_dir ),
                        dot( v3, ray_dir ) );

            d03 = vec4( b03.x * b03.x - dot( v0, v0 ) + obj_pos[i+0].w,
                        b03.y * b03.y - dot( v1, v1 ) + obj_pos[i+1].w,
                        b03.z * b03.z - dot( v2, v2 ) + obj_pos[i+2].w,
                        b03.w * b03.w - dot( v3, v3 ) + obj_pos[i+3].w );

            dv = greaterThan( d03, vec4( 0.0f ) );

            if( any( dv ) ) {

                vec4 t03 = -b03 - sqrt( d03 );

                dv = dv && greaterThan( t03, vec4( 0.0f ) );

                if( dv.x && t03.x < rt_dist ) { rt_dist = t03.x; rt_hit = i+0; }
                if( dv.y && t03.y < rt_dist ) { rt_dist = t03.y; rt_hit = i+1; }
                if( dv.z && t03.z < rt_dist ) { rt_dist = t03.z; rt_hit = i+2; }
                if( dv.w && t03.w < rt_dist ) { rt_dist = t03.w; rt_hit = i+3; }

            }

            // 4-7
            v0 = ray_pos - obj_pos[i+4].xyz;
            v1 = ray_pos - obj_pos[i+5].xyz;
            v2 = ray_pos - obj_pos[i+6].xyz;
            v3 = ray_pos - obj_pos[i+7].xyz;

            b03 = vec4( dot( v0, ray_dir ),
                        dot( v1, ray_dir ),
                        dot( v2, ray_dir ),
                        dot( v3, ray_dir ) );

            d03 = vec4( b03.x * b03.x - dot( v0, v0 ) + obj_pos[i+4].w,
                        b03.y * b03.y - dot( v1, v1 ) + obj_pos[i+5].w,
                        b03.z * b03.z - dot( v2, v2 ) + obj_pos[i+6].w,
                        b03.w * b03.w - dot( v3, v3 ) + obj_pos[i+7].w );

            dv = greaterThan( d03, vec4( 0.0f ) );

            if( any( dv ) ) {

                vec4 t03 = -b03 - sqrt( d03 );

                dv = dv && greaterThan( t03, vec4( 0.0f ) );

                if( dv.x && t03.x < rt_dist ) { rt_dist = t03.x; rt_hit = i+4; }
                if( dv.y && t03.y < rt_dist ) { rt_dist = t03.y; rt_hit = i+5; }
                if( dv.z && t03.z < rt_dist ) { rt_dist = t03.z; rt_hit = i+6; }
                if( dv.w && t03.w < rt_dist ) { rt_dist = t03.w; rt_hit = i+7; }

            }

            // 8-11
            v0 = ray_pos - obj_pos[i+ 8].xyz;
            v1 = ray_pos - obj_pos[i+ 9].xyz;
            v2 = ray_pos - obj_pos[i+10].xyz;
            v3 = ray_pos - obj_pos[i+11].xyz;

            b03 = vec4( dot( v0, ray_dir ),
                        dot( v1, ray_dir ),
                        dot( v2, ray_dir ),
                        dot( v3, ray_dir ) );

            d03 = vec4( b03.x * b03.x - dot( v0, v0 ) + obj_pos[i+ 8].w,
                        b03.y * b03.y - dot( v1, v1 ) + obj_pos[i+ 9].w,
                        b03.z * b03.z - dot( v2, v2 ) + obj_pos[i+10].w,
                        b03.w * b03.w - dot( v3, v3 ) + obj_pos[i+11].w );

            dv = greaterThan( d03, vec4( 0.0f ) );

            if( any( dv ) ) {

                vec4 t03 = -b03 - sqrt( d03 );

                dv = dv && greaterThan( t03, vec4( 0.0f ) );

                if( dv.x && t03.x < rt_dist ) { rt_dist = t03.x; rt_hit = i+ 8; }
                if( dv.y && t03.y < rt_dist ) { rt_dist = t03.y; rt_hit = i+ 9; }
                if( dv.z && t03.z < rt_dist ) { rt_dist = t03.z; rt_hit = i+10; }
                if( dv.w && t03.w < rt_dist ) { rt_dist = t03.w; rt_hit = i+11; }

            }

            // 12-15
            v0 = ray_pos - obj_pos[i+12].xyz;
            v1 = ray_pos - obj_pos[i+13].xyz;
            v2 = ray_pos - obj_pos[i+14].xyz;
            v3 = ray_pos - obj_pos[i+15].xyz;

            b03 = vec4( dot( v0, ray_dir ),
                        dot( v1, ray_dir ),
                        dot( v2, ray_dir ),
                        dot( v3, ray_dir ) );

            d03 = vec4( b03.x * b03.x - dot( v0, v0 ) + obj_pos[i+12].w,
                        b03.y * b03.y - dot( v1, v1 ) + obj_pos[i+13].w,
                        b03.z * b03.z - dot( v2, v2 ) + obj_pos[i+14].w,
                        b03.w * b03.w - dot( v3, v3 ) + obj_pos[i+15].w );

            dv = greaterThan( d03, vec4( 0.0f ) );

            if( any( dv ) ) {

                vec4 t03 = -b03 - sqrt( d03 );

                dv = dv && greaterThan( t03, vec4( 0.0f ) );

                if( dv.x && t03.x < rt_dist ) { rt_dist = t03.x; rt_hit = i+12; }
                if( dv.y && t03.y < rt_dist ) { rt_dist = t03.y; rt_hit = i+13; }
                if( dv.z && t03.z < rt_dist ) { rt_dist = t03.z; rt_hit = i+14; }
                if( dv.w && t03.w < rt_dist ) { rt_dist = t03.w; rt_hit = i+15; }

            }

        }

        if( rt_hit == -1 && refl > 0 ) break;

        // plane with shadows
        if( rt_hit == -1 && refl == 0 ) {
            rt_dist = -(dot( ray_pos, plane_norm.xyz ) + plane_norm.w) / dot( ray_dir, plane_norm.xyz );

            if( rt_dist <= 0.0f ) {
                // plane miss, background
                vec2 dxy = v_texCoord - 0.5f;
                float dv = 1.0f - sqrt( dot( dxy, dxy ) + (sin(obj_pos[1].z)+1.0f)/2.0f );
                col = plane_col.yzx * dv;
                break;
            }

            //  plane hit
            ray_pos += ray_dir * rt_dist;
            vec3 l   = normalize( light_pos.xyz - ray_pos );

            // check shadow
            int i;
            for( i = 0; i < OBJNUM; i += 16 ) {
                vec3 v0, v1, v2, v3;
                vec4 b03, d03;

                v0 = ray_pos - obj_pos[i+0].xyz;
                v1 = ray_pos - obj_pos[i+1].xyz;
                v2 = ray_pos - obj_pos[i+2].xyz;
                v3 = ray_pos - obj_pos[i+3].xyz;

                b03 = vec4( dot( v0, l ),
                            dot( v1, l ),
                            dot( v2, l ),
                            dot( v3, l ) );

                d03 = vec4( b03.x * b03.x - dot( v0, v0 ) + obj_pos[i+0].w,
                            b03.y * b03.y - dot( v1, v1 ) + obj_pos[i+1].w,
                            b03.z * b03.z - dot( v2, v2 ) + obj_pos[i+2].w,
                            b03.w * b03.w - dot( v3, v3 ) + obj_pos[i+3].w );

                if( any( greaterThan( d03, vec4( 0.0f ) ) ) ) break;

                v0 = ray_pos - obj_pos[i+4].xyz;
                v1 = ray_pos - obj_pos[i+5].xyz;
                v2 = ray_pos - obj_pos[i+6].xyz;
                v3 = ray_pos - obj_pos[i+7].xyz;

                b03 = vec4( dot( v0, l ),
                            dot( v1, l ),
                            dot( v2, l ),
                            dot( v3, l ) );

                d03 = vec4( b03.x * b03.x - dot( v0, v0 ) + obj_pos[i+4].w,
                            b03.y * b03.y - dot( v1, v1 ) + obj_pos[i+5].w,
                            b03.z * b03.z - dot( v2, v2 ) + obj_pos[i+6].w,
                            b03.w * b03.w - dot( v3, v3 ) + obj_pos[i+7].w );

                if( any( greaterThan( d03, vec4( 0.0f ) ) ) ) break;

                v0 = ray_pos - obj_pos[i+ 8].xyz;
                v1 = ray_pos - obj_pos[i+ 9].xyz;
                v2 = ray_pos - obj_pos[i+10].xyz;
                v3 = ray_pos - obj_pos[i+11].xyz;

                b03 = vec4( dot( v0, l ),
                            dot( v1, l ),
                            dot( v2, l ),
                            dot( v3, l ) );

                d03 = vec4( b03.x * b03.x - dot( v0, v0 ) + obj_pos[i+ 8].w,
                            b03.y * b03.y - dot( v1, v1 ) + obj_pos[i+ 9].w,
                            b03.z * b03.z - dot( v2, v2 ) + obj_pos[i+10].w,
                            b03.w * b03.w - dot( v3, v3 ) + obj_pos[i+11].w );

                if( any( greaterThan( d03, vec4( 0.0f ) ) ) ) break;

                v0 = ray_pos - obj_pos[i+12].xyz;
                v1 = ray_pos - obj_pos[i+13].xyz;
                v2 = ray_pos - obj_pos[i+14].xyz;
                v3 = ray_pos - obj_pos[i+15].xyz;

                b03 = vec4( dot( v0, l ),
                            dot( v1, l ),
                            dot( v2, l ),
                            dot( v3, l ) );

                d03 = vec4( b03.x * b03.x - dot( v0, v0 ) + obj_pos[i+12].w,
                            b03.y * b03.y - dot( v1, v1 ) + obj_pos[i+13].w,
                            b03.z * b03.z - dot( v2, v2 ) + obj_pos[i+14].w,
                            b03.w * b03.w - dot( v3, v3 ) + obj_pos[i+15].w );

                if( any( greaterThan( d03, vec4( 0.0f ) ) ) ) break;

            }

            float diffuse = i == OBJNUM ? dot( plane_norm.xyz, l ) * 1.2f : 0.0f;
            col = plane_col.xyz * diffuse;
            break;

        }

        // sphere hit + reflection
        ray_pos += ray_dir * rt_dist;
        vec3 n   = normalize( ray_pos - obj_pos[rt_hit].xyz );
        vec3 l   = normalize( light_pos.xyz - ray_pos );

        float diffuse  = max( dot( n, l ), 0.0f );
        float specular = pow( max( dot( ray_dir, l - n * 2.0f * diffuse ), 0.0f ), 8.0f ); // ! always pos diffuse

        col += light_col.xyz * specular + obj_col[rt_hit].xyz * diffuse;

        ray_dir = reflect( ray_dir, n );

    }

    outColor = vec4( col, 1.0f );

}
